package common.enchantment;

import java.util.List;
import java.util.Map;
import java.util.Set;

import common.collect.Lists;
import common.collect.Maps;
import common.entity.DamageSource;
import common.entity.Entity;
import common.entity.types.EntityLiving;
import common.item.ItemArmor;
import common.item.ItemAxe;
import common.item.ItemShears;
import common.item.ItemStack;
import common.rng.Random;
import common.util.Displayable;
import common.util.ExtMath;
import common.util.Identifyable;
import common.vars.Vars;

public enum Enchantment implements Displayable, Identifyable
{   
    PROTECTION(0, "protection", 10, EnumEnchantmentType.ARMOR, "Schutz") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 1 + (enchantmentLevel - 1) * 11;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 20;
        }

        public int getMaxLevel()
        {
            return 4;
        }

        public int calcDamageReduction(int level, DamageSource source)
        {
            if (source == DamageSource.outOfWorld)
            {
                return 0;
            }
            else
            {
                float f = (float)(6 + level * level) / 3.0F;
                return ExtMath.floorf(f * 0.75F);
            }
        }
    },
    FIRE_PROTECTION(1, "fire_protection", 5, EnumEnchantmentType.ARMOR, "Feuerschutz") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 10 + (enchantmentLevel - 1) * 8;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 12;
        }

        public int getMaxLevel()
        {
            return 4;
        }

        public int calcDamageReduction(int level, DamageSource source)
        {
            if (source == DamageSource.outOfWorld)
            {
                return 0;
            }
            else
            {
                float f = (float)(6 + level * level) / 3.0F;
                return source.isFireDamage() ? ExtMath.floorf(f * 1.25F) : 0;
            }
        }
    },
    FEATHER_FALLING(2, "feather_falling", 5, EnumEnchantmentType.ARMOR_FEET, "Federfall") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 5 + (enchantmentLevel - 1) * 6;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 10;
        }

        public int getMaxLevel()
        {
            return 4;
        }

        public int calcDamageReduction(int level, DamageSource source)
        {
            if (source == DamageSource.outOfWorld)
            {
                return 0;
            }
            else
            {
                float f = (float)(6 + level * level) / 3.0F;
                return source == DamageSource.fall ? ExtMath.floorf(f * 2.5F) : 0;
            }
        }
    },
    BLAST_PROTECTION(3, "blast_protection", 2, EnumEnchantmentType.ARMOR, "Explosionsschutz") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 5 + (enchantmentLevel - 1) * 8;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 12;
        }

        public int getMaxLevel()
        {
            return 4;
        }

        public int calcDamageReduction(int level, DamageSource source)
        {
            if (source == DamageSource.outOfWorld)
            {
                return 0;
            }
            else
            {
                float f = (float)(6 + level * level) / 3.0F;
                return source.isExplosion() ? ExtMath.floorf(f * 1.5F) : 0;
            }
        }
    },
    PROJECTILE_PROTECTION(4, "projectile_protection", 5, EnumEnchantmentType.ARMOR, "Schusssicher") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 3 + (enchantmentLevel - 1) * 6;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 15;
        }

        public int getMaxLevel()
        {
            return 4;
        }

        public int calcDamageReduction(int level, DamageSource source)
        {
            if (source == DamageSource.outOfWorld)
            {
                return 0;
            }
            else
            {
                float f = (float)(6 + level * level) / 3.0F;
                return source.isProjectile() ? ExtMath.floorf(f * 1.5F) : 0;
            }
        }
    },
    THORNS(7, "thorns", 1, EnumEnchantmentType.ARMOR_TORSO, "Dornen") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 10 + 20 * (enchantmentLevel - 1);
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }

        public boolean canApply(ItemStack stack)
        {
            return stack.getItem() instanceof ItemArmor ? true : super.canApply(stack);
        }

        public void onUserHurt(EntityLiving user, Entity attacker, int level)
        {
            Random random = user.getRNG();
            ItemStack itemstack = EnchantmentHelper.getEnchantedItem(Enchantment.THORNS, user);

            if ((user.worldObj.client || Vars.damageThorns) && isThornsHurting(level, random))
            {
                if (attacker != null)
                {
                    attacker.attackEntityFrom(DamageSource.causeThornsDamage(user), getThornsDamage(level, random));
//                    attacker.playSound("damage.thorns", 0.5F, 1.0F);
                }

                if (itemstack != null)
                {
                    itemstack.damageItem(3, user);
                }
            }
            else if (itemstack != null)
            {
                itemstack.damageItem(1, user);
            }
        }
    },
    SHARPNESS(16, "sharpness", 10, EnumEnchantmentType.WEAPON, "Schärfe") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 1 + (enchantmentLevel - 1) * 11;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 20;
        }

        public int getMaxLevel()
        {
            return 5;
        }

        public int calcAdditionalDamage(int level)
        {
            return (int)(/*this.damageType == 0 ? */ (float)level * 1.25F); // : (this.damageType == 1 && creatureType == CreatureType.UNDEAD ? (float)level * 2.5F : (this.damageType == 2 && creatureType == CreatureType.ARTHROPOD ? (float)level * 2.5F : 0.0F)));
        }

        public boolean canApply(ItemStack stack)
        {
            return stack.getItem() instanceof ItemAxe ? true : super.canApply(stack);
        }

//        /**
//         * Called whenever a mob is damaged with an item that has this enchantment on it.
//         */
//        public void onEntityDamaged(EntityLivingBase user, Entity target, int level)
//        {
//            if (target instanceof EntityLivingBase)
//            {
//                EntityLivingBase entitylivingbase = (EntityLivingBase)target;
    //
//                if (this.damageType == 2 && entitylivingbase.getCreatureType() == CreatureType.ARTHROPOD)
//                {
//                    int i = 20 + user.getRNG().zrange(10 * level);
//                    entitylivingbase.addEffect(new PotionEffect(Potion.moveSlowdown.id, i, 3));
//                }
//            }
//        }
    },
    KNOCKBACK(19, "knockback", 5, EnumEnchantmentType.WEAPON, "Rückstoß") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 5 + 20 * (enchantmentLevel - 1);
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 2;
        }
    },
    FIRE_ASPECT(20, "fire_aspect", 2, EnumEnchantmentType.WEAPON, "Verbrennung") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 10 + 20 * (enchantmentLevel - 1);
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 2;
        }
    },
    LOOTING(21, "looting", 2, EnumEnchantmentType.WEAPON, "Plünderung") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 15 + (enchantmentLevel - 1) * 9;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }

        public boolean canApplyTogether(Enchantment ench)
        {
            return super.canApplyTogether(ench) && ench.id != SILK_TOUCH.id;
        }
    },
    EFFICIENCY(32, "efficiency", 10, EnumEnchantmentType.DIGGER, "Effizienz") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 1 + 10 * (enchantmentLevel - 1);
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 5;
        }

        public boolean canApply(ItemStack stack)
        {
            return stack.getItem() instanceof ItemShears ? true : super.canApply(stack);
        }
    },
    SILK_TOUCH(33, "silk_touch", 1, EnumEnchantmentType.DIGGER, "Behutsamkeit") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 15;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 1;
        }

        public boolean canApplyTogether(Enchantment ench)
        {
            return super.canApplyTogether(ench) && ench.id != FORTUNE.id;
        }

        public boolean canApply(ItemStack stack)
        {
            return stack.getItem() instanceof ItemShears ? true : super.canApply(stack);
        }
    },
    UNBREAKING(34, "unbreaking", 5, EnumEnchantmentType.BREAKABLE, "Haltbarkeit") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 5 + (enchantmentLevel - 1) * 8;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }

        public boolean canApply(ItemStack stack)
        {
            return stack.isItemStackDamageable() ? true : super.canApply(stack);
        }
    },
    FORTUNE(35, "fortune", 2, EnumEnchantmentType.DIGGER, "Glück") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 15 + (enchantmentLevel - 1) * 9;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }

        public boolean canApplyTogether(Enchantment ench)
        {
            return super.canApplyTogether(ench) && ench.id != SILK_TOUCH.id;
        }
    },
    POWER(48, "power", 10, EnumEnchantmentType.BOW, "Stärke") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 1 + (enchantmentLevel - 1) * 10;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 15;
        }

        public int getMaxLevel()
        {
            return 5;
        }
    },
    PUNCH(49, "punch", 2, EnumEnchantmentType.BOW, "Schlag") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 12 + (enchantmentLevel - 1) * 20;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return this.getMinEnchantability(enchantmentLevel) + 25;
        }

        public int getMaxLevel()
        {
            return 2;
        }
    },
    FLAME(50, "flame", 2, EnumEnchantmentType.BOW, "Flamme") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 20;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return 50;
        }

        public int getMaxLevel()
        {
            return 1;
        }
    },
    INFINITY(51, "infinity", 1, EnumEnchantmentType.BOW, "Unendlich") {
    	public int getMinEnchantability(int enchantmentLevel)
        {
            return 20;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return 50;
        }

        public int getMaxLevel()
        {
            return 1;
        }
    },
    LUCK_OF_THE_SEA(61, "luck_of_the_sea", 2, EnumEnchantmentType.FISHING_ROD, "Glück des Meeres") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 15 + (enchantmentLevel - 1) * 9;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }

        public boolean canApplyTogether(Enchantment ench)
        {
            return super.canApplyTogether(ench) && ench.id != SILK_TOUCH.id;
        }
    },
    LURE(62, "lure", 2, EnumEnchantmentType.FISHING_ROD, "Köder") {
        public int getMinEnchantability(int enchantmentLevel)
        {
            return 15 + (enchantmentLevel - 1) * 9;
        }

        public int getMaxEnchantability(int enchantmentLevel)
        {
            return super.getMinEnchantability(enchantmentLevel) + 50;
        }

        public int getMaxLevel()
        {
            return 3;
        }
    },
    DRAINING(64, "draining", 1, EnumEnchantmentType.WEAPON, "Seelenentzug") {
    	public int getMinEnchantability(int enchantmentLevel) {
    		return 2 + (enchantmentLevel - 1) * 11;
    	}

    	public int getMaxEnchantability(int enchantmentLevel) {
    		return this.getMinEnchantability(enchantmentLevel) + 20;
    	}

    	public int getMaxLevel() {
    		return 16;
    	}

    	public boolean canApply(ItemStack stack) {
    		return stack.getItem() instanceof ItemAxe ? true : super.canApply(stack);
    	}

    	public void onEntityDamaged(EntityLiving user, Entity target, int level) {
    		if(target instanceof EntityLiving) {
    			EntityLiving entity = (EntityLiving)target;
    			entity.attackEntityFrom(DamageSource.causeIndirectMagicDamage(user, user), (int)((float)level * 1.25F));
    			user.heal((int)((entity.arePotionsInverted() ? 0.2f : 1.0f) * (float)level));
    		}
    	}
    };
    
    private static final String[] LEVELS = new String[] {"I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X"};
    private static final Enchantment[] LOOKUP = new Enchantment[256];
    public static final Enchantment[] ENCHANTMENTS;
    private static final Map<String, Enchantment> NAME_LOOKUP = Maps.<String, Enchantment>newHashMap();
    
    public final int id;
    private final String name;
    private final int weight;
    public final EnumEnchantmentType type;
    private final String display;
    
    static {
    	for(Enchantment ench : values()) {
            if(LOOKUP[ench.id] != null)
                throw new IllegalArgumentException("Verzauberung mit id " + ench.id + " ist bereits registriert");
            if(NAME_LOOKUP.get(ench.name) != null)
                throw new IllegalArgumentException("Verzauberung mit name " + ench.name + " ist bereits registriert");
            LOOKUP[ench.id] = ench;
            NAME_LOOKUP.put(ench.name, ench);
    	}
        List<Enchantment> list = Lists.<Enchantment>newArrayList();
        for(Enchantment enchantment : LOOKUP) {
            if(enchantment != null)
                list.add(enchantment);
        }
        ENCHANTMENTS = list.toArray(new Enchantment[list.size()]);
    }

    public static Enchantment getEnchantmentById(int enchID)
    {
        return enchID >= 0 && enchID < LOOKUP.length ? LOOKUP[enchID] : null;
    }

    public static Enchantment getEnchantmentByLocation(String location)
    {
        return NAME_LOOKUP.get(location);
    }

    public static Set<String> getNames()
    {
        return NAME_LOOKUP.keySet();
    }

    private Enchantment(int enchID, String enchName, int enchWeight, EnumEnchantmentType enchType, String display)
    {
        this.id = enchID;
        this.name = enchName;
        this.weight = enchWeight;
        this.type = enchType;
        this.display = display;
    }

    public int getWeight()
    {
        return this.weight;
    }

    public int getMinLevel()
    {
        return 1;
    }

    public int getMaxLevel()
    {
        return 1;
    }

    public int getMinEnchantability(int enchantmentLevel)
    {
        return 1 + enchantmentLevel * 10;
    }

    public int getMaxEnchantability(int enchantmentLevel)
    {
        return this.getMinEnchantability(enchantmentLevel) + 5;
    }

    public int calcDamageReduction(int level, DamageSource source)
    {
        return 0;
    }

    public int calcAdditionalDamage(int level)
    {
        return 0;
    }

    public boolean canApplyTogether(Enchantment ench)
    {
        return this != ench;
    }
    
    public final String getName() {
    	return this.name;
    }

    public String getDisplay()
    {
        return this.display;
    }
    
    public String getFormattedName(int level)
    {
        return this.getDisplay() + " " + ((level >= 1 && level <= LEVELS.length) ? LEVELS[level - 1] : "" + level);
    }

    public boolean canApply(ItemStack stack)
    {
        return this.type.canEnchantItem(stack.getItem());
    }

    public void onEntityDamaged(EntityLiving user, Entity target, int level)
    {
    }

    public void onUserHurt(EntityLiving user, Entity attacker, int level)
    {
    }

    public static int getFireTimeForEntity(Entity entity, int duration)
    {
        int i = EnchantmentHelper.getMaxEnchantmentLevel(Enchantment.FIRE_PROTECTION.id, entity.getInventory());

        if (i > 0)
        {
            duration -= ExtMath.floorf((float)duration * (float)i * 0.15F);
        }

        return duration;
    }

    public static double getKnockbackFactor(Entity entity, double strength)
    {
        int i = EnchantmentHelper.getMaxEnchantmentLevel(Enchantment.BLAST_PROTECTION.id, entity.getInventory());

        if (i > 0)
        {
            strength -= (double)ExtMath.floord(strength * (double)((float)i * 0.15F));
        }

        return strength;
    }

    public static boolean negateDamage(ItemStack stack, int level, Random rand)
    {
        return stack.getItem() instanceof ItemArmor && rand.floatv() < 0.6F ? false : rand.rarity(level + 1);
    }

    public static boolean isThornsHurting(int level, Random rand)
    {
        return level <= 0 ? false : rand.floatv() < 0.15F * (float)level;
    }

    public static int getThornsDamage(int level, Random rand)
    {
        return level > 10 ? level - 10 : rand.roll(4);
    }
}
